__all__ = (
    "init_main_process_logger",
    "init_child_process_logger",
    "get_logger",
)

import atexit
import logging
import multiprocessing
import os
from logging.handlers import (
    QueueHandler,
    QueueListener,
)
from typing import (
    List,
    Tuple,
)

from neptune_scale.core.styles import (
    STYLES,
    ensure_style_detected,
)
from neptune_scale.envs import DEBUG_MODE

NEPTUNE_LOGGER_NAME = "neptune"
NEPTUNE_DEBUG_FILE_NAME = "neptune.log"
LOG_FORMAT = "{blue}%(name)s{end} :: {bold}%(levelname)s{end} :: %(message)s"
DEBUG_FORMAT = "%(asctime)s :: %(name)s :: %(levelname)s :: %(processName)s(%(process)d):%(threadName)s:%(filename)s:%(funcName)s():%(lineno)d %(message)s"


def get_logger() -> logging.Logger:
    """Use in modules to get the root Neptune logger"""
    return logging.getLogger(NEPTUNE_LOGGER_NAME)


def init_main_process_logger() -> Tuple[logging.Logger, multiprocessing.Queue]:
    """
    Initialize the root 'neptune' logger to use a mp.Queue. This should be called only once in the main process.

    Returns:
        A 2-tuple of (logger, queue). Queue should be passed to `init_child_logger()` in child processes.
    """

    if multiprocessing.parent_process() is not None:
        raise RuntimeError("This function should be called only in the main process.")

    logger = logging.getLogger(NEPTUNE_LOGGER_NAME)

    # If the user has also imported `neptune-fetcher` the root logger will already be initialized.
    # We want our handlers to take precedence. We will remove all handlers and add our own.
    if logger.hasHandlers():
        # Not initialized by us, clear handlers and proceed.
        if not hasattr(logger, "__neptune_scale_listener"):
            logger.handlers.clear()
        else:
            return logger, _get_queue_from_handlers(logger.handlers)

    logger.setLevel(logging.INFO)
    ensure_style_detected()

    handlers: List[logging.Handler] = []
    if os.environ.get(DEBUG_MODE, "False").lower() in ("true", "1"):
        logger.setLevel(logging.DEBUG)

        file_handler = logging.FileHandler(NEPTUNE_DEBUG_FILE_NAME)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(DEBUG_FORMAT))
        handlers.append(file_handler)

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(LOG_FORMAT.format(**STYLES)))
    handlers.append(stream_handler)

    queue: multiprocessing.Queue = multiprocessing.Queue()
    logger.addHandler(QueueHandler(queue))

    listener = QueueListener(queue, *handlers, respect_handler_level=True)
    listener.start()

    logger.__neptune_scale_listener = listener  # type: ignore
    atexit.register(listener.stop)

    return logger, queue


def init_child_process_logger(queue: multiprocessing.Queue) -> logging.Logger:
    """
    Initialize a child logger to use the given queue. This should be called in child processes only once.
    After it's called, the logger can be retrieved using `get_logger()`.

    Args:
        queue: A multiprocessing.Queue object returned by `init_root_logger()`.

    Returns:
        A logger instance.
    """

    if multiprocessing.parent_process() is None:
        raise RuntimeError("This function should be called only in child processes.")

    logger = logging.getLogger(NEPTUNE_LOGGER_NAME)
    if logger.hasHandlers():
        # Make sure the QueueHandler is already registered
        _ = _get_queue_from_handlers(logger.handlers)
        return logger

    ensure_style_detected()

    logger.addHandler(QueueHandler(queue))
    logger.setLevel(logging.INFO)

    return logger


def _get_queue_from_handlers(handlers: List[logging.Handler]) -> multiprocessing.Queue:
    for h in handlers:
        if isinstance(h, QueueHandler):
            return h.queue  # type: ignore # mypy doesn't know it's always Queue

    raise RuntimeError("Expected to find a QueueHandler in the logger handlers.")



from __future__ import annotations

__all__ = ("MetadataSplitter",)

from datetime import datetime
from itertools import starmap
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from more_itertools import peekable
from neptune_api.proto.neptune_pb.ingest.v1.common_pb2 import (
    SET_OPERATION,
    UpdateRunSnapshot,
    Value,
)
from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation

from neptune_scale.core.serialization import (
    datetime_to_proto,
    make_step,
    make_value,
    pb_key_size,
)

T = TypeVar("T", bound=Any)


class MetadataSplitter(Iterator[Tuple[RunOperation, int]]):
    def __init__(
        self,
        *,
        project: str,
        run_id: str,
        step: Optional[Union[int, float]],
        timestamp: datetime,
        fields: Dict[str, Union[float, bool, int, str, datetime, list, set, tuple]],
        metrics: Dict[str, float],
        add_tags: Dict[str, Union[List[str], Set[str], Tuple[str]]],
        remove_tags: Dict[str, Union[List[str], Set[str], Tuple[str]]],
        max_message_bytes_size: int = 1024 * 1024,
    ):
        self._step = None if step is None else make_step(number=step)
        self._timestamp = datetime_to_proto(timestamp)
        self._project = project
        self._run_id = run_id
        self._fields = peekable(fields.items())
        self._metrics = peekable(starmap(lambda k, v: (k, float(v)), metrics.items()))
        self._add_tags = peekable(add_tags.items())
        self._remove_tags = peekable(remove_tags.items())

        self._max_update_bytes_size = (
            max_message_bytes_size
            - RunOperation(
                project=self._project,
                run_id=self._run_id,
                update=UpdateRunSnapshot(step=self._step, timestamp=self._timestamp),
            ).ByteSize()
        )

        self._has_returned = False

    def __iter__(self) -> MetadataSplitter:
        self._has_returned = False
        return self

    def __next__(self) -> Tuple[RunOperation, int]:
        update = UpdateRunSnapshot(
            step=self._step,
            timestamp=self._timestamp,
            assign={},
            append={},
            modify_sets={},
        )
        size = update.ByteSize()

        size = self.populate(
            assets=self._fields,
            update_producer=lambda key, value: update.assign[key].MergeFrom(value),
            size=size,
        )
        size = self.populate(
            assets=self._metrics,
            update_producer=lambda key, value: update.append[key].MergeFrom(value),
            size=size,
        )
        size = self.populate_tags(
            update=update,
            assets=self._add_tags,
            operation=SET_OPERATION.ADD,
            size=size,
        )
        size = self.populate_tags(
            update=update,
            assets=self._remove_tags,
            operation=SET_OPERATION.REMOVE,
            size=size,
        )

        if not self._has_returned or update.assign or update.append or update.modify_sets:
            self._has_returned = True
            return RunOperation(project=self._project, run_id=self._run_id, update=update), size
        else:
            raise StopIteration

    def populate(
        self,
        assets: peekable[Any],
        update_producer: Callable[[str, Value], None],
        size: int,
    ) -> int:
        while size < self._max_update_bytes_size:
            try:
                key, value = assets.peek()
            except StopIteration:
                break

            proto_value = make_value(value)
            new_size = size + pb_key_size(key) + proto_value.ByteSize() + 6

            if new_size > self._max_update_bytes_size:
                break

            update_producer(key, proto_value)
            size, _ = new_size, next(assets)

        return size

    def populate_tags(
        self, update: UpdateRunSnapshot, assets: peekable[Any], operation: SET_OPERATION.ValueType, size: int
    ) -> int:
        while size < self._max_update_bytes_size:
            try:
                key, values = assets.peek()
            except StopIteration:
                break

            if not isinstance(values, peekable):
                values = peekable(values)

            is_full = False
            new_size = size + pb_key_size(key) + 6
            for value in values:
                tag_size = pb_key_size(value) + 6
                if new_size + tag_size > self._max_update_bytes_size:
                    values.prepend(value)
                    is_full = True
                    break

                update.modify_sets[key].string.values[value] = operation
                new_size += tag_size

            size, _ = new_size, next(assets)

            if is_full:
                assets.prepend((key, list(values)))
                break

        return size



__all__ = ("STYLES", "ensure_style_detected")

import os
import platform
from typing import Dict

from neptune_scale.envs import DISABLE_COLORS

UNIX_STYLES = {
    "h1": "\033[95m",
    "h2": "\033[94m",
    "blue": "\033[94m",
    "python": "\033[96m",
    "bash": "\033[95m",
    "warning": "\033[93m",
    "correct": "\033[92m",
    "fail": "\033[91m",
    "bold": "\033[1m",
    "underline": "\033[4m",
    "end": "\033[0m",
}

WINDOWS_STYLES = {
    "h1": "",
    "h2": "",
    "blue": "",
    "python": "",
    "bash": "",
    "warning": "",
    "correct": "",
    "fail": "",
    "bold": "",
    "underline": "",
    "end": "",
}

EMPTY_STYLES = {
    "h1": "",
    "h2": "",
    "blue": "",
    "python": "",
    "bash": "",
    "warning": "",
    "correct": "",
    "fail": "",
    "bold": "",
    "underline": "",
    "end": "",
}


STYLES: Dict[str, str] = {}


def ensure_style_detected() -> None:
    if not STYLES:
        if os.environ.get(DISABLE_COLORS, "False").lower() in ("true", "1"):
            STYLES.update(EMPTY_STYLES)
        else:
            if platform.system() in ["Linux", "Darwin"]:
                STYLES.update(UNIX_STYLES)
            elif platform.system() == "Windows":
                STYLES.update(WINDOWS_STYLES)
            else:
                STYLES.update(EMPTY_STYLES)



from __future__ import annotations

__all__ = ("AggregatingQueue",)

import time
from queue import (
    Empty,
    Queue,
)
from threading import RLock
from typing import Optional

from neptune_api.proto.neptune_pb.ingest.v1.pub.ingest_pb2 import RunOperation

from neptune_scale.core.components.abstract import Resource
from neptune_scale.core.components.queue_element import (
    BatchedOperations,
    SingleOperation,
)
from neptune_scale.core.logger import get_logger
from neptune_scale.parameters import (
    BATCH_WAIT_TIME_SECONDS,
    MAX_BATCH_SIZE,
    MAX_QUEUE_ELEMENT_SIZE,
)

logger = get_logger()


class AggregatingQueue(Resource):
    def __init__(
        self,
        max_queue_size: int,
        max_elements_in_batch: int = MAX_BATCH_SIZE,
        max_queue_element_size: int = MAX_QUEUE_ELEMENT_SIZE,
        wait_time: float = BATCH_WAIT_TIME_SECONDS,
    ) -> None:
        self._max_queue_size = max_queue_size
        self._max_elements_in_batch = max_elements_in_batch
        self._max_queue_element_size = max_queue_element_size
        self._wait_time = wait_time

        self._queue: Queue[SingleOperation] = Queue(maxsize=max_queue_size)
        self._lock: RLock = RLock()
        self._latest_unprocessed: Optional[SingleOperation] = None

    @property
    def maxsize(self) -> int:
        return self._max_queue_size

    def put_nowait(self, element: SingleOperation) -> None:
        with self._lock:
            self._queue.put_nowait(element)

    def _get_next(self, timeout: float) -> Optional[SingleOperation]:
        # We can assume that each of queue elements are less than MAX_QUEUE_ELEMENT_SIZE because of MetadataSplitter.
        # We can assume that every queue element has the same project, run id and family
        with self._lock:
            if self._latest_unprocessed is not None:
                return self._latest_unprocessed

            try:
                self._latest_unprocessed = self._queue.get(timeout=timeout)
                return self._latest_unprocessed
            except Empty:
                return None

    def commit(self) -> None:
        self._latest_unprocessed = None

    def get(self) -> BatchedOperations:
        start = time.monotonic()

        batch_operations: list[RunOperation] = []
        last_batch_key: Optional[float] = None
        batch_sequence_id: Optional[int] = None
        batch_timestamp: Optional[float] = None

        batch_bytes: int = 0
        elements_in_batch: int = 0
        wait_remaining = self._wait_time

        # Pull operations off the queue until we either reach the maximum size, or
        # the specified wait time has passed. This way we maximize the potential batch size.
        while True:
            t0 = time.monotonic()

            if elements_in_batch >= self._max_elements_in_batch:
                logger.debug("Batch closed due to limit of elements in batch %s", elements_in_batch)
                break

            element = self._get_next(wait_remaining)
            if element is None:
                break

            if not batch_operations or element.batch_key != last_batch_key:
                new_operation = RunOperation()
                new_operation.ParseFromString(element.operation)
                batch_operations.append(new_operation)
                last_batch_key = element.batch_key
            else:
                if not element.is_batchable:
                    logger.debug("Batch closed due to next operation not being batchable")
                    break

                assert element.metadata_size is not None  # mypy, metadata update always has metadata size

                if batch_bytes + element.metadata_size > self._max_queue_element_size:
                    logger.debug("Batch closed due to size limit %s", batch_bytes + element.metadata_size)
                    break

                new_operation = RunOperation()
                new_operation.ParseFromString(element.operation)
                merge_run_operation(batch_operations[-1], new_operation)

            batch_sequence_id = element.sequence_id
            batch_timestamp = element.timestamp

            if element.metadata_size is not None:
                batch_bytes += element.metadata_size
            else:
                batch_bytes += len(element.operation)
            elements_in_batch += 1

            self.commit()

            if not element.is_batchable:
                logger.debug("Batch closed due to first not being batchable")
                break

            t1 = time.monotonic()
            wait_remaining -= t1 - t0

            if wait_remaining <= 0:
                logger.debug("Batch closed due to wait time")
                break

        if not batch_operations:
            logger.debug(f"Batch is empty after {self._wait_time} seconds of waiting.")
            raise Empty

        assert batch_sequence_id is not None  # mypy
        assert batch_timestamp is not None  # mypy

        logger.debug(
            "Batched %d operations. Total size %d. Total time %f",
            elements_in_batch,
            batch_bytes,
            time.monotonic() - start,
        )

        batch = create_run_batch(batch_operations)

        return BatchedOperations(
            sequence_id=batch_sequence_id,
            timestamp=batch_timestamp,
            operation=batch.SerializeToString(),
        )


def create_run_batch(operations: list[RunOperation]) -> RunOperation:
    if len(operations) == 1:
        return operations[0]

    batch = RunOperation()

    head_operation = operations[0]
    batch.project = head_operation.project
    batch.run_id = head_operation.run_id
    batch.create_missing_project = head_operation.create_missing_project
    batch.api_key = head_operation.api_key

    for operation in operations:
        operation_type = operation.WhichOneof("operation")
        if operation_type == "update":
            batch.update_batch.snapshots.append(operation.update)
        else:
            raise ValueError("Cannot batch operation of type %s", operation_type)

    return batch


def merge_run_operation(batch: RunOperation, operation: RunOperation) -> None:
    """
    Merge the `operation` into `batch`, taking into account the special case of `modify_sets`.

    Protobuf merges existing map keys by simply overwriting values, instead of calling
    `MergeFrom` on the existing value, eg: A['foo'] = B['foo'].

    We want this instead:

        batch = {'sys/tags': 'string': { 'values': {'foo': ADD}}}
        operation = {'sys/tags': 'string': { 'values': {'bar': ADD}}}
        result = {'sys/tags': 'string': { 'values': {'foo': ADD, 'bar': ADD}}}

    If we called `batch.MergeFrom(operation)` we would get an overwritten value:
        result = {'sys/tags': 'string': { 'values': {'bar': ADD}}}

    This function ensures that the `modify_sets` are merged correctly, leaving the default
    behaviour for all other fields.
    """

    modify_sets = operation.update.modify_sets
    operation.update.ClearField("modify_sets")

    batch.MergeFrom(operation)

    for k, v in modify_sets.items():
        batch.update.modify_sets[k].MergeFrom(v)



# Input validation
MAX_RUN_ID_LENGTH = 128
MAX_EXPERIMENT_NAME_LENGTH = 730

# Operations queue
MAX_BATCH_SIZE = 100000
MAX_QUEUE_SIZE = 1000000
MAX_MULTIPROCESSING_QUEUE_SIZE = 32767
MAX_QUEUE_ELEMENT_SIZE = 1024 * 1024  # 1MB
# Wait up to this many seconds for incoming operations before submitting a batch
BATCH_WAIT_TIME_SECONDS = 1

# Threads
SYNC_THREAD_SLEEP_TIME = 2
STATUS_TRACKING_THREAD_SLEEP_TIME = 1
INTERNAL_QUEUE_FEEDER_THREAD_SLEEP_TIME = 0.01
ERRORS_MONITOR_THREAD_SLEEP_TIME = 0.1
SYNC_PROCESS_SLEEP_TIME = 1
LAG_TRACKER_THREAD_SLEEP_TIME = 1

# Networking
REQUEST_TIMEOUT = 5
MAX_REQUEST_RETRY_SECONDS = 60

# User facing
SHUTDOWN_TIMEOUT = 60  # 1 minute
MINIMAL_WAIT_FOR_PUT_SLEEP_TIME = 10
MINIMAL_WAIT_FOR_ACK_SLEEP_TIME = 10
STOP_MESSAGE_FREQUENCY = 5
LAG_TRACKER_TIMEOUT = 1

# Status tracking
MAX_REQUESTS_STATUS_BATCH_SIZE = 1000
